enum
{
	coa_color = 0x0F,
	coa_type = 0xF0,
	
	coa_none = 0,
	
	coa_white = 0x2,
	coa_black = 0x3,
	
	coa_pawn = 0x10,
	coa_knight = 0x20,
	coa_bishop = 0x30,
	coa_rook = 0x40,
	coa_queen = 0x50,
	coa_king = 0x60,
};

struct coa_board
{
	unsigned char pieces[64];
	unsigned char depth;
	unsigned long long int count;
};

#define inline static inline __attribute((always_inline))

static void coa_perft(struct coa_board *board, int turn);

inline int coa_play(struct coa_board *board, int turn, int x0, int y0, int x1, int y1)
{
	if (x1 < 0) return 1;
	if (y1 < 0) return 1;
	if (x1 >= 8) return 1;
	if (y1 >= 8) return 1;
	
	unsigned char piece = board->pieces[x0 + y0 * 8];
	unsigned char other = board->pieces[x1 + y1 * 8];
	if ((other & coa_color) == (piece & coa_color)) return 1;
	
	board->pieces[x0 + y0 * 8] = coa_none;
	board->pieces[x1 + y1 * 8] = piece;
	coa_perft(board, turn ^ 1);
	board->pieces[x0 + y0 * 8] = piece;
	board->pieces[x1 + y1 * 8] = other;
	
	if (other != coa_none) return 1;
	return 0;
}

inline void coa_move_slider(struct coa_board *board, int turn, int x0, int y0, int dx, int dy)
{
	int x1 = x0;
	int y1 = y0;
	
	do
	{
		x1 += dx;
		y1 += dy;
	}
	while (coa_play(board, turn, x0, y0, x1, y1) == 0);
}

inline void coa_move_pawn(struct coa_board *board, int turn, int x0, int y0)
{
	int dx, sy;
	if (turn == coa_white)
		dx = 1, sy = 1;
	else
		dx = -1, sy = 6;
	
	int y1 = y0 + dx;
	
	if (y1 == 0 || y1 == 7)
	{
		board->pieces[x0 + y0 * 8] &= coa_color;
		board->pieces[x0 + y0 * 8] |= coa_queen;
	}
	
	if (board->pieces[x0 + y1 * 8] == coa_none)
	{
		coa_play(board, turn, x0, y0, x0, y1);
		
		if (y0 == sy)
		{
			int y2 = y1 + dx;
			if (board->pieces[x0 + y2 * 8] == coa_none)
				coa_play(board, turn, x0, y0, x0, y2);
		}
	}
	
	int xl = x0 - 1;
	int xr = x0 + 1;
	
	if (board->pieces[xl + y1 * 8])
		coa_play(board, turn, x0, y0, xl, y1);
	if (board->pieces[xr + y1 * 8])
		coa_play(board, turn, x0, y0, xr, y1);
	
	if (y1 == 0 || y1 == 7)
	{
		board->pieces[x0 + y0 * 8] &= coa_color;
		board->pieces[x0 + y0 * 8] |= coa_pawn;
	}
}

inline void coa_move_knight(struct coa_board *board, int turn, int x0, int y0)
{
	coa_play(board, turn, x0, y0, x0 + 1, y0 + 2);
	coa_play(board, turn, x0, y0, x0 + 2, y0 + 1);
	coa_play(board, turn, x0, y0, x0 - 1, y0 + 2);
	coa_play(board, turn, x0, y0, x0 - 2, y0 + 1);
	coa_play(board, turn, x0, y0, x0 + 1, y0 - 2);
	coa_play(board, turn, x0, y0, x0 - 2, y0 + 1);
	coa_play(board, turn, x0, y0, x0 - 1, y0 - 2);
	coa_play(board, turn, x0, y0, x0 - 2, y0 - 1);
}

inline void coa_move_bishop(struct coa_board *board, int turn, int x0, int y0)
{
	coa_move_slider(board, turn, x0, y0, -1, -1);
	coa_move_slider(board, turn, x0, y0, -1, 1);
	coa_move_slider(board, turn, x0, y0, 1, -1);
	coa_move_slider(board, turn, x0, y0, 1, 1);
}

inline void coa_move_rook(struct coa_board *board, int turn, int x0, int y0)
{
	coa_move_slider(board, turn, x0, y0, 0, -1);
	coa_move_slider(board, turn, x0, y0, -1, 0);
	coa_move_slider(board, turn, x0, y0, 0, 1);
	coa_move_slider(board, turn, x0, y0, 1, 0);
}

inline void coa_move_queen(struct coa_board *board, int turn, int x0, int y0)
{
	coa_move_rook(board, turn, x0, y0);
	coa_move_bishop(board, turn, x0, y0);
}

inline void coa_move_king(struct coa_board *board, int turn, int x0, int y0)
{
	coa_play(board, turn, x0, y0, x0 - 1, y0 - 1);
	coa_play(board, turn, x0, y0, x0 - 1, y0 + 0);
	coa_play(board, turn, x0, y0, x0 - 1, y0 + 1);
	
	coa_play(board, turn, x0, y0, x0 + 0, y0 - 1);
	coa_play(board, turn, x0, y0, x0 + 0, y0 + 0);
	coa_play(board, turn, x0, y0, x0 + 0, y0 + 1);
	
	coa_play(board, turn, x0, y0, x0 + 1, y0 - 1);
	coa_play(board, turn, x0, y0, x0 + 1, y0 + 0);
	coa_play(board, turn, x0, y0, x0 + 1, y0 + 1);
}

inline void coa_move(struct coa_board *board, int turn, int x0, int y0)
{
	unsigned char piece = board->pieces[x0 + y0 * 8];
	if ((piece & coa_color) != turn) return;
	switch (piece & coa_type)
	{
		case coa_pawn: coa_move_pawn(board, turn, x0, y0); break;
		case coa_knight: coa_move_knight(board, turn, x0, y0); break;
		case coa_bishop: coa_move_bishop(board, turn, x0, y0); break;
		case coa_rook: coa_move_rook(board, turn, x0, y0); break;
		case coa_queen: coa_move_queen(board, turn, x0, y0); break;
		case coa_king: coa_move_king(board, turn, x0, y0); break;
	}
}

static void coa_perft(struct coa_board *board, int turn)
{
	if (board->depth == 0)
	{
		board->count++;
		return;
	}
	
	board->depth--;
	
	#pragma GCC unroll 100
	for (int y = 0 ; y < 8 ; y++)
	#pragma GCC unroll 100
	for (int x = 0 ; x < 8 ; x++)
		coa_move(board, turn, x, y);
	
	board->depth++;
}

#include <stdio.h>
#include <time.h>

int main(void)
{
	struct coa_board board =
	{
		{
			coa_white|coa_rook, coa_white|coa_knight, coa_white|coa_bishop, coa_white|coa_queen, coa_white|coa_king, coa_white|coa_bishop, coa_white|coa_knight, coa_white|coa_rook,
			coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn, coa_white|coa_pawn,
			coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none,
			coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none,
			coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none,
			coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none, coa_none,
			coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn, coa_black|coa_pawn,
			coa_black|coa_rook, coa_black|coa_knight, coa_black|coa_bishop, coa_black|coa_queen, coa_black|coa_king, coa_black|coa_bishop, coa_black|coa_knight, coa_black|coa_rook,
		},
		0, 0,
	};
	
	for (int i = 0 ; i < 8 ; i++)
	{
		board.depth = i;
		board.count = 0;
		clock_t t0 = clock();
		coa_perft(&board, coa_white);
		double t = clock() - t0;
		
		printf("depth %2d -> ", i);
		if (board.count == 1)
			printf("%10lld node  / ", board.count);
		else
			printf("%10lld nodes / ", board.count);
		
		if (t / 1000 < 10)
			printf("%4.f ns     = ", t);
		else if (t / 1000 / 1000 < 10)
			printf("%8.3f ms = ", t / 1000);
		else
			printf("%8.3f  s = ", t / 1000 / 1000);
		
		printf("%7.3f Mn/s\n", board.count / t);
	}
	
	return 0;
}
